import shutil
from pathlib import Path
import json
import subprocess


def _make_spec(path_to_python, name):
    return {
        "argv": [
            str(path_to_python), "-m", "ipykernel_launcher", "-f",
            "{connection_file}"
        ],
        "display_name":
        f"Python 3 ({name})",
        "language":
        "python",
        "metadata": {
            "debugger": True
        }
    }


def _get_info():
    from conda.cli.python_api import run_command, Commands

    # maybe replace this with sys.prefix?
    out, err, code = run_command(Commands.INFO, "--json")
    return json.loads(out)


def _current_prefix():
    return _get_info()["active_prefix"]


def _install(name, requirements, inline=False, verbose=False):
    """Create a new conda environment
    """
    from mamba import api

    requirements = list(requirements) + ["ipykernel"]

    fn = api.install if inline else api.create

    if verbose:
        fn(
            name,
            requirements,
            ('conda-forge', ),
        )

    else:
        # NOTE: capturing output when using the Python API isn't working
        cmds = ['install'] if inline else ['create', '-n', name]
        cmd = ['mamba'] + cmds + requirements + ['-c', 'conda-forge', '-y']
        subprocess.run(cmd, capture_output=True, check=True)


def install(name, requirements, inline=False, verbose=False):
    """Create a new kernel
    """
    prefix = _current_prefix()

    # delete existing environment, otherwise mamba will complain with a
    # cryptic error
    prefix_new = Path(prefix, 'envs', name)

    if prefix_new.is_dir():
        shutil.rmtree(prefix_new)

    # create conda environment
    _install(name, requirements, inline=inline, verbose=verbose)

    # newly created environment
    path_to_python = Path(prefix, 'envs', name, 'bin', 'python')

    # folder where we'll register the kernel (in the prefix)
    path_to_kernels = Path(prefix, 'share', 'jupyter', 'kernels')

    spec = _make_spec(path_to_python, name)

    path_to_kernel = Path(path_to_kernels, name)
    path_to_kernel.mkdir(exist_ok=True)

    (path_to_kernel / 'kernel.json').write_text(json.dumps(spec))



import re
import sys
from pathlib import Path

import parso
from isort import place_module

from k2s.env import install


def extract_from_plain_text(text):
    matches = re.findall(r'pip install ([\w \-]+)', text)
    deps = set([
        item for sublist in [match.split() for match in matches]
        for item in sublist if not item.startswith('-')
    ])
    return list(deps - {'k2s'})


# NOTE: we copied this from soorgeon
# https://github.com/ploomber/soorgeon/blob/cb16dc51f326cf21486bff86dfec03b4d1f5c2d7/src/soorgeon/definitions.py#L24
def packages_used(tree):
    """
    Return a list of the packages used, correcting for some packages whose
    module name does not match the PyPI package (e.g., sklearn -> scikit-learn)
    Returns None if fails to parse them
    """

    pkg_name = {
        'sklearn': 'scikit-learn',
        'sql': 'jupysql',
    }

    def flatten(elements):
        return [i for sub in elements for i in sub]

    def _extract_names(node):
        if hasattr(node, 'children'):
            return extract_names(node.children[0])
        else:
            return [node.value]

    def extract_names(import_):
        if import_.type == 'name':
            return [import_.value]
        elif import_.type in {'dotted_name', 'dotted_as_name'}:
            return [import_.children[0].value]

        second = import_.children[1]

        if second.type in {'dotted_name', 'dotted_as_name'}:
            return extract_names(second.children[0])
        elif second.type == 'dotted_as_names':
            # import a as something, b as another

            return flatten([
                _extract_names(node) for i, node in enumerate(second.children)
                if i % 2 == 0
            ])
        else:
            return [second.value]

    pkgs = flatten([extract_names(import_) for import_ in tree.iter_imports()])

    # replace using pkg_name mapping and ignore standard lib
    pkgs_final = [
        pkg_name.get(name, name) for name in pkgs
        if place_module(name) == 'THIRDPARTY'
    ]

    # remove duplicates and sort
    return sorted(set(pkgs_final))


def extract_imports(nb):

    try:
        code = '\n'.join([
            cell['source'] for cell in nb['cells']
            if cell['cell_type'] == "code"
        ])
    except TypeError:
        # if passed a notebook read using json instead of nbformat
        code = '\n'.join([
            '\n'.join(c['source']) for c in nb['cells']
            if c['cell_type'] == 'code'
        ])

    try:
        plain = '\n'.join([cell['source'] for cell in nb['cells']])
    except TypeError:
        # # if passed a notebook read using json instead of nbformat
        plain = '\n'.join(['\n'.join(c['source']) for c in nb['cells']])

    return packages_used(parso.parse(code)) + extract_from_plain_text(plain)


def bootstrap_env(path_to_notebook, inline=False, verbose=False):
    import nbformat

    if inline:
        name = Path(sys.prefix).name
    else:
        name = Path(path_to_notebook).stem

    print('Parsing notebook...')
    nb = nbformat.read(path_to_notebook, as_version=nbformat.NO_CONVERT)
    imports = extract_imports(nb)
    imports = set(imports) - {'k2s'}
    imports_str = ', '.join(imports)
    print(f'Found: {imports_str}. Installing...')

    install(name, imports, verbose=verbose, inline=inline)

    if not inline:
        print('Kernel is ready! Refresh your browser')

    nb.metadata.kernelspec = {
        'display_name': f'Python 3 ({name})',
        'language': 'python',
        'name': name,
    }

    # override kernel spec - note that refreshing won't switch to this
    # kernel and it'll stick with the original one until we kill the existing
    # kernel
    Path(path_to_notebook).write_text(nbformat.v4.writes(nb))



"""
pip install k2s --upgrade && k2s get ploomber/jupysql/doc/nb.ipynb
"""
import os
import json
import sys
from pathlib import Path, PurePosixPath
import subprocess
import venv
import urllib.request

from k2s.parse import extract_imports


def _get_python_folder_and_bin_name():
    folder = 'Scripts' if os.name == 'nt' else 'bin'
    bin_name = 'python.exe' if os.name == 'nt' else 'python'
    return folder, bin_name


def from_url(url):
    """

    >>> from k2s.bootstrap import from_url; from_url("URL")
    """
    url = f'https://raw.githubusercontent.com/{url}'

    file = PurePosixPath(url).name
    urllib.request.urlretrieve(url, file)
    return from_file(file)


def from_file(file):
    env_name = Path(file).stem + '-env'

    nb = json.loads(Path(file).read_text(encoding='utf-8'))

    deps = extract_imports(nb)

    print("Creating virtual environment...")
    print(f"Installing: {', '.join(deps)}")
    print("This may take a moment. "
          "Join our community while you wait: https://ploomber.io/community")

    # maybe site-packages, yes? will that copy the existing ones? perhaps
    # they already ahve most of the dependencies and we can save time.
    # if so, then we need to ensure jupyterlab starts with the kernel in
    # the virtual environment
    venv.create(env_name, with_pip=True, system_site_packages=False)

    dir_, bin_name = _get_python_folder_and_bin_name()
    path_to_python = str(Path(env_name, dir_, bin_name))

    subprocess.run(
        [path_to_python, '-m', 'pip', 'install', 'pip', '--upgrade'],
        check=True,
        capture_output=True)

    subprocess.run(
        [path_to_python, '-m', 'pip', 'install', 'jupyterlab', '--quiet'] +
        deps,
        check=True,
        capture_output=False)

    print("Lauching Jupyter...")

    if Path(f'{env_name}/bin/jupyter-lab').exists():
        jupyterlab = f'{env_name}/bin/jupyter-lab'
    else:
        jupyterlab = str(Path(sys.prefix, 'bin', 'jupyter-lab'))

    jupyterlab = f'{env_name}/bin/jupyter-lab'

    print('To exit: CTRL + C')

    try:
        subprocess.run([jupyterlab, file], check=True, capture_output=True)
    except KeyboardInterrupt:
        print('Exiting...')
        print(f'\nTo laucnh again:\n\n'
              f'{env_name}/bin/jupyter-lab {file}')
